/*********************************
 *MON: A Curses Memory Monitor
 *Brandon Lucia 9/20/05
 *Tufts University
 *Comp111 - Operating Systems
 *********************************
 Since this application makes use 
 of the curses library for fancy
 textual output, you will need to 
 compile it with the following gcc
 call:
 gcc -lm -lcurses mon.c -o mon
 -lm, since we're linking to math
 -lcurses since we're liking to curses
 *********************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <pwd.h>
#include <sys/types.h>
#include <dirent.h>
#include <signal.h>
#include <string.h>
#include <curses.h>
#define MAX_PROCESSES 65536
#define LEN_HISTORY 100 
#define ANOMALOUS 1 
#define NORMAL 0 
#undef DEBUG

struct process_info{
	char user[16];
	char cmd_line[48];
	int mem_increase;
	int last_size;
	int alive;
};

/*************************************************************
 *these are the dreaded, but necessary globals
 *They hold the process data, the process states (dead/alive)
 *and they keep track of which spot in history we are
 *. they need to be global so that signal handlers
 *can sack their data if they want to do so in an asynchronous
 * way
 *************************************************************/
struct process_info **Process_Data_List;
//struct process_info *Process_Data_List[MAX_PROCESSES];
//int lhd[LEN_HISTORY];
int *lhd;

/***************************
 *SIGNAL HANDLERS LIVE HERE*
 ***************************/
//Does what it says - dumps the differences to a file
//this is the sighandler registered for SIGINT
void dump_diffs(int sig){
	
	int i;	
	FILE *diffs_file;
	clear();
	if((diffs_file = fopen("mon.diffs","w"))){
		for(i=0;i<LEN_HISTORY;i++)
			fprintf(diffs_file,"%d\n",lhd[i]);
		fclose(diffs_file);
	}
	else
		printf("Warning: Couldn't Save Statistics to file %s\n","mon.diffs");

		exit(0);
}

void print_detailed_mem_usage(int sig){
	int i;
	int count=0;
	int total_mem=0;
	clear();
	move(0,0);
	attron(A_BOLD);
	attron(A_REVERSE);
	printw("                         MON:Curses Memory Monitor                             ");
	attroff(A_REVERSE);
	attroff(A_BOLD);
	move(2,0);
	attron(A_BOLD);
	printw("---------------------Detailed Memory Report Below---------------------------");
	attroff(A_BOLD);
	move(3,0);
	attron(A_REVERSE);
	printw("PID    MemUse  Command                                                      ");
	attroff(A_REVERSE);
	for(i=0; i<MAX_PROCESSES; i++){
		if(Process_Data_List[i]){
			if(count+4<LINES){
				move(count+4,0);
				printw("%d",i);
				move(count+4,7);
				printw("%d",Process_Data_List[i]->last_size);
				move(count+4,14);
				printw("(%s)",Process_Data_List[i]->cmd_line);
			}
			count++;	
			total_mem+=Process_Data_List[i]->last_size;
		}
	}
	move(1,0);
	attron(A_BOLD);
	printw("Total Data/Stack Use: ");
	attroff(A_BOLD);
	printw("%dkB\n",total_mem);
	refresh();
	sleep(3);
}


//prints the data/stack usage
void print_DS_usage(int used){
	clear();
	move(0,0);
	attron(A_BOLD);
	attron(A_REVERSE);
	printw("                         MON:Curses Memory Monitor                              ");
	attroff(A_REVERSE);
	move(1,0);
	printw("Total Data/Stack Use: ");
	attroff(A_BOLD);
	printw("%dkB\n",used);
	refresh();
}

//bubble sort.....on the difference value (anomalous output)
void sort_process_list_of_n_by_diff(int n,struct process_info **sort_list){
	int j=0;
	int k=0;
	for(j=0;j<n-1;j++){
		for(k=0;k<n-1-j;k++){
			if(sort_list[k] && sort_list[k+1]){
				if(sort_list[k+1]->mem_increase > 
				   sort_list[k]->mem_increase){
					struct process_info *tmp = sort_list[k];
					sort_list[k] = sort_list[k+1];
					sort_list[k+1]=tmp;
				}
			}	
		}
	}
}

//bubble sort.....on the total memsize for this process (normal output)
void sort_process_list_of_n_by_size(int n,struct process_info **sort_list){
	int j=0;
	int k=0;
	for(j=0;j<n-1;j++){
		for(k=0;k<n-1-j;k++){
			if(sort_list[k] && sort_list[k+1]){
				if(sort_list[k+1]->last_size > 
				   sort_list[k]->last_size){
					struct process_info *tmp = sort_list[k];
					sort_list[k] = sort_list[k+1];
					sort_list[k+1]=tmp;
				}
			}	
		}
	}
}

//Prints the data from a process_info struct in a nice way
void print_process_info(int rank,struct process_info *this_proc){
	if(rank+4 < LINES){
		move(rank+4,4);
		printw("%s",this_proc->user);
		move(rank+4,18);
		printw("%d",this_proc->last_size);
		move(rank+4,27);
		printw("%d",this_proc->mem_increase);
		move(rank+4,36);
		printw("%s",this_proc->cmd_line);
	}
}

//prints a "root cause analysis" report in a nice way
void print_root_cause_analysis(int anomaly,struct process_info **ProcList){	
	if(anomaly){
		attron(A_REVERSE);
		attron(A_BOLD);
		move(2,0);
		printw("--Memory Anomaly Detected--------(Potential Offenders Listed Below)-------------");
		attroff(A_BOLD);
		attroff(A_REVERSE);
	}
	attron(A_REVERSE);
	move(3,0);
	printw("USER              MemUse   MemInc   Command                                     ");
	attroff(A_REVERSE);
	int how_many=0;
	int j=0;
	struct process_info *sort_list[MAX_PROCESSES];
	for(j=0; j<MAX_PROCESSES;j++){
		if(ProcList[j]){
			if(anomaly){
				if(ProcList[j]->mem_increase>0)
					sort_list[how_many++]=ProcList[j];
			}
			else
				sort_list[how_many++]=ProcList[j];
		}
	}
	//print sorted by diff if anomalous
	if(anomaly)
		sort_process_list_of_n_by_diff(how_many,sort_list);
	else
		sort_process_list_of_n_by_size(how_many,sort_list);
	//print sorted by size if normal
		
	for(j=0; j<how_many; j++){
		print_process_info(j,sort_list[j]);
	}
	refresh();

}




///Calculates the standard deviation of n point in points
double standard_deviation(int *points,double n){
	double sum=0;
	double sum_of_squares=0;
	double std_dev=0;
	int i;
	for(i=0; i<LEN_HISTORY; i++){
		sum+=(double)points[i];
		sum_of_squares+=((double)points[i]*(double)points[i]);
	}
	double variance = 
	(1/(((double)LEN_HISTORY)-1))  *  
	(sum_of_squares - ((1/((((double)LEN_HISTORY)-1)))  *  (sum*sum))  );
	if(variance>=0)
		std_dev = sqrt(variance);
	return std_dev;
}

//calculate the mean of n points stored in points
double mean(int * points,int n){
	double sum=0;
	double mean=0;
	int i;
	for(i=0; i<LEN_HISTORY; i++){
		sum+=(double)points[i];
	}
	mean = sum/((double)LEN_HISTORY)-1;
	return mean;
}
		
void parse_into_proc_list_at_n(char *nstr){
	int n=atoi(nstr);	
	char current_line[256];
	char current_label[256];
	char file_path[256];
	int data_size=0;
	int stk_size=0;
	FILE *proc_file;

	//THE MONITOR
	/***************************************
	* There are two cases:
	* 1) We don't have a process with this pid (all new stuff)
	* 2) We already have a process with this pid (just update the memdif)
	***************************************/
	//1) NEW PROCESS
	if(!Process_Data_List[n]){
		//Create a new entry in the processes hash
		Process_Data_List[n] = (struct process_info *) malloc(
			                sizeof(struct process_info));
	
		//add the cmd_line of this process to the object 
		sprintf(file_path,"/proc/%s/cmdline",nstr);
		if((proc_file = fopen(file_path,"r"))){
			char cmd[512];
			fscanf(proc_file,"%s",cmd);
			strncpy(Process_Data_List[n]->cmd_line,cmd,47);
			Process_Data_List[n]->cmd_line[47]='\0';
			fclose(proc_file);
		}
		else{
			strncpy(Process_Data_List[n]->cmd_line,"<none>",47);
		}

		sprintf(file_path,"/proc/%s/status",nstr);
		if((proc_file = fopen(file_path,"r"))){
			while(fgets(current_line,256,proc_file)){
				sscanf(current_line,"%s",current_label);
				//Get the UID of the executor	
				if(!strcmp(current_label,"Uid:")){
					char label[16];
					int uid=0;
					struct passwd *this_user;
					sscanf(current_line,"%s %d",
					       label,&uid);
					this_user=getpwuid(uid);
					strncpy(Process_Data_List[n]->user,
						this_user->pw_name,15);
				}
				//get the memory stuff for the first time
				if(!strcmp(current_label,"VmData:")){
					char label[16];
					sscanf(current_line,
					       "%s %d",
					       label,&data_size);
				}
				if(!strcmp(current_label,"VmStk:")){
					char label[16];
					sscanf(current_line,
					       "%s %d",
					       label,&stk_size);
				}
			}
			Process_Data_List[n]->mem_increase=data_size+stk_size;
			Process_Data_List[n]->last_size=data_size+stk_size;
			fclose(proc_file);
			Process_Data_List[n]->alive=1;
		}
		else{
			printf("couldn't open proc filefor process %d\n",n);
			Process_Data_List[n]->mem_increase=0;
			Process_Data_List[n]->last_size=0;
			Process_Data_List[n]->alive=0;
		}
	}
	//2) This process has been seen before...
	else{
		sprintf(file_path,"/proc/%s/status",nstr);
		if((proc_file = fopen(file_path,"r"))){
			while(fgets(current_line,256,proc_file)){
				sscanf(current_line,"%s",current_label);
				if(!strcmp(current_label,"VmData:")){
					char label[128];
					sscanf(current_line,
					       "%s %d",
					       label,&data_size);
				}
				if(!strcmp(current_label,"VmStk:")){
					char label[128];
					sscanf(current_line,
					       "%s %d",
					       label,&stk_size);
				}
			}
			//Update the memory increase based on the last size	
			Process_Data_List[n]->mem_increase=
				(data_size+stk_size)-
				Process_Data_List[n]->last_size;
			//Update the last size	
			Process_Data_List[n]->last_size=
				(data_size+stk_size);

			fclose(proc_file);
			Process_Data_List[n]->alive=1;
		}
		else{
			printf("couldn't open proc file for process %d\n",n);
			Process_Data_List[n]->mem_increase=0;
			Process_Data_List[n]->last_size=0;
			Process_Data_List[n]->alive=0;
		}
	}
}

/***********************************************************
 *This is surely the meat of the monitor, but there's really
 *not much to it.  This updates the global proc list, and  
 *if appropriate issues a root_cause_analysis report.
 *Reading of proc is taken care of in this function as well.
 **********************************************************/
int run_mon(int* last_hundred,int which_iter,int last_size){
	DIR * proc_handle;
	struct dirent *d;
	int total_used=0;
	int i;
	//assume all processes are gone, then when we see them, call them alive
	for(i=0; i<MAX_PROCESSES; i++){
		if(Process_Data_List[i])
			Process_Data_List[i]->alive=0;
	}
	
if((proc_handle = opendir("/proc"))){
	while((d = readdir(proc_handle))){
		//dirs that are #'s only	
		if(d->d_name[0]<='9' && d->d_name[0]>='0'){
			parse_into_proc_list_at_n(d->d_name);	
		}
	}
	closedir(proc_handle);
}
else{
	printf("Error: Couldn't Open /proc...Quitting!\n\n");
	exit(0);
}
	//here, if a process that was here last time is gone,
	//get rid of it, since process id's can be reused
	for(i=0;i<MAX_PROCESSES; i++){
		if(Process_Data_List[i] && !Process_Data_List[i]->alive){
			free(Process_Data_List[i]);
			Process_Data_List[i]=NULL;
		}
		//accumulate the total used memory
		if(Process_Data_List[i]){
			total_used+=Process_Data_List[i]->last_size;
		}
	}
		print_DS_usage(total_used);
		last_hundred[which_iter]=total_used-last_size;

	//if there is an offensively large increase in memory use...
	double this_mean = mean(last_hundred,which_iter);
	double this_std_dev = standard_deviation(last_hundred,which_iter);
	if((((double)last_hundred[which_iter])>
	    (this_mean+(2*this_std_dev))) 
	  && 
	   ((this_mean+(2*this_std_dev))>0)
	  ){
	  	//ANOMALY!
	  	print_root_cause_analysis(ANOMALOUS,Process_Data_List);
	   }
	else
	  	print_root_cause_analysis(NORMAL,Process_Data_List);
		//NORMAL...
		
return total_used;
}
//need to allocate and zero the process hash
struct process_info **initialize_process_hash(){
	struct process_info **p;
	if((p = (struct process_info **) malloc(sizeof(struct process_info *)*MAX_PROCESSES))){
		bzero(p,(MAX_PROCESSES*sizeof(struct process_info *)));	
		return p;
	}
	else{
		printf("Fatal Error: Couldn't Allocate Process List!\n");
		exit(1);
	}
}

//need to allocate and zero the difference history
int *initialize_difference_history(){
	int *h;
	if((h = (int *) malloc(sizeof(int *)*LEN_HISTORY))){
		bzero(h,(LEN_HISTORY*sizeof(int *)));	
		return h;
	}else{
		printf("Fatal Error: Couldn't Allocate History Record List!\n");
		exit(1);
	}
}
		
int main(){
	FILE *diffs_file;
	Process_Data_List=initialize_process_hash();
	lhd=initialize_difference_history();
	if((diffs_file = fopen("mon.diffs","r"))){
		int q=0;
		char cur_ln[256];
		while(fgets(cur_ln,256,diffs_file)){
			sscanf(cur_ln,"%d\n",&lhd[q]);
			q++;
		}
	}
	
	struct sigaction INT_action;
	INT_action.sa_handler = dump_diffs;
	sigfillset(&INT_action.sa_mask);
	sigaction(SIGTERM,&INT_action,NULL);
	
	struct sigaction HUP_action;
	HUP_action.sa_handler = print_detailed_mem_usage;
	sigfillset(&HUP_action.sa_mask);
	sigaction(SIGHUP,&HUP_action,NULL);
	
	sigset_t blocked;
	sigset_t old;
	sigfillset(&blocked);
	
	int iter=0;
	int last_size=0;
	initscr();
	while(1){
		//we start touching global things, so lets block signals
		sigprocmask(SIG_BLOCK,&blocked,&old);
		last_size = run_mon(lhd,iter,last_size);	
		iter = (iter+1)%99;
		sigprocmask(SIG_UNBLOCK,&blocked,&old);
		sleep(5);
	}
	return 0;
}
